/*  This file is part of portdowngrade.

    portdowngrade is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    portdowngrade is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with portdowngrade; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#define VERSION_HI 0
#define VERSION_LO 6

#include "../config.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>
#include <dirent.h>
#include <termios.h>
#include <fcntl.h>
#ifndef TIOCGWINSIZE
#include <sys/ioctl.h>
#endif

using namespace std;

#include <popt.h>
#include <string>
#include <fstream>
#include <list>
#include <iostream>
#include <vector>
#include <algorithm>

typedef list<string> StringList;

string portsDir = "/usr/ports";
string serverRootDir = "";
string defaultServerRootDir = DEFAULT_SERVER;
string portname;
int MaxTimeDiff = 10;
int MinCommentLength = 256;
ofstream logStream;

class TempDir
{
	string i_tempDir;
public:
	TempDir()
	{
		string tempDir = tmpnam(0);
		tempDir += "/";
		if (mkdir(tempDir.c_str(), S_IRWXU))
			perror("portupgrade error: can not create temporary directory");
		else
			i_tempDir = tempDir;
	}

	bool tempDirExists() const {return i_tempDir.length() > 0;}

	const string & getTempPath() const {return i_tempDir;}

	string getTempFile() const {return i_tempDir + "temp";}

	~TempDir()
	{
		if (tempDirExists())
		{
			string command = "rm -rfd " + i_tempDir;
			int i = system(command.c_str());
			if (!WIFEXITED(i) || WEXITSTATUS(i))
				fprintf(stderr, "portdowngrade error: Can not delete temporary directory %s\n", i_tempDir.c_str());
		}
	}
} tempDir;

typedef vector<class CVSCommitEvent> CommitEventVectorType;

class CVSCommitEvent
{
	string i_date, i_comment, i_fileName, i_portVersion;
	bool i_portVersionSet;
	time_t i_unixDate;
	void setDate(const string & date)
	{
		i_date = date;
		struct tm t;
		memset(&t, 0, sizeof(t));
		char * p = strptime(date.c_str(), "%Y/%m/%d %H:%M:%S", &t);
		if (p && !*p)
			i_unixDate = mktime(&t);
		else
			fprintf(stderr, "date format error: %s\n", date.c_str());
	}
public:
	CVSCommitEvent() : i_unixDate(0), i_portVersionSet(false) {}
	CVSCommitEvent(const string & date, const string & comment, const string & fileName)
		: i_unixDate(0), i_portVersionSet(false), i_comment(comment), i_fileName(fileName) {setDate(date);}
	CVSCommitEvent(const CVSCommitEvent & src) : i_unixDate(0) {operator=(src);}
	const CVSCommitEvent & operator=(const CVSCommitEvent & src)
		{i_date = src.i_date; i_unixDate = src.i_unixDate; i_comment = src.i_comment; i_fileName = src.i_fileName;
		i_portVersion = src.i_portVersion; i_portVersionSet = src.i_portVersionSet;}
	const string & getDate() const {return i_date;}
	time_t getUnixDate() const {return i_unixDate;}
	time_t diff(const CVSCommitEvent & event) const {return getUnixDate() - event.getUnixDate();}
	const string & getComment() const {return i_comment;}
	const string & getFileName() const {return i_fileName;}
	bool isMakefile() const {return getFileName() == "Makefile";}
	void addToVector(CommitEventVectorType & vector) const {vector.push_back(*this);}
	void setPortVersion(const string & portVersion) {i_portVersion = portVersion; i_portVersionSet = true;}
	bool isPortVersionSet() const {return i_portVersionSet;}
	const string & getPortVersion() const {return i_portVersion;}
};

CommitEventVectorType CommitEvents;

struct CommitEventsLargerThan
{
	bool operator()(const CVSCommitEvent & e1, const CVSCommitEvent & e2)
	{
		return e1.getUnixDate() > e2.getUnixDate();
	}
};

istream& operator<<(string & s, istream & stream)
{
	s = "";
	char buffer[1024], * p;
	if (stream.fail() || stream.eof())
		return stream;
	do
	{
		*buffer = 0;
		p = buffer;
		do
		{
			stream.get(*p++);
			if (*(p-1) == '\r')
				stream.get(*(p-1));
		}
		while (!stream.fail() && !stream.eof() && (p - buffer < sizeof(buffer)) && *(p-1) != '\n');
		if (*(p-1) == '\n')
			*(p-1) = 0;
		*p = 0;
		s += buffer;
		if (!*(p-1))
			break;
	}
	while (!stream.fail() && !stream.eof());
	return stream;
}

// http://www.erlenstar.demon.co.uk/unix/faq_4.html

static struct termios stored_settings;

void storeTermiosSettings()
{
	tcgetattr(0,&stored_settings);
}

void restoreTermiosSettings()
{
	tcsetattr(0,TCSANOW,&stored_settings);
}

void echo_off(void)
{
	struct termios new_settings;
	tcgetattr(0,&new_settings);
	new_settings.c_lflag &= (~ECHO);
	tcsetattr(0,TCSANOW,&new_settings);
}

void set_keypress(void)
{
	struct termios new_settings;

	tcgetattr(0,&new_settings);

	/* Disable canonical mode, and set buffer size to 1 byte */
	new_settings.c_lflag &= (~ICANON);
	new_settings.c_cc[VTIME] = 0;
	new_settings.c_cc[VMIN] = 1;

	tcsetattr(0,TCSANOW,&new_settings);
}

int kbhit()
{
	int Input;
	int flags;
	int Result;

	flags = fcntl( STDIN_FILENO , F_GETFL);

	/* set stdin to NON-BLOCKING */
	fcntl( STDIN_FILENO , F_SETFL , flags | O_NONBLOCK );
	Input = getc(stdin);
	if ( Input != -1 ) {
		ungetc( Input, stdin );
		Result = 1;

	}
	else
		Result =0;

	/* put things pack how we found them */

	fcntl ( STDIN_FILENO , F_SETFL , flags );
	return ( Result );
}


bool completePortPath(string & portname)
{
	vector<string> portNames;
	fprintf(stdout, "Seeking port %s ... ", portname.c_str());
	string command = "cd " + portsDir + "; make search name=";
	int i1 = portname.find('/'), i2 = portname.find('$');
	if (i1 >= 0 || i2 >= 0)
	{
		string s = portname;
		if (i1 >= 0)
			s.erase(0, i1+1);
		if (i2 >= 0)
			s.erase(i2-i1-1);
		command += s + " | grep ^Path: | grep " + portname + " > " + tempDir.getTempFile();
	}
	else
		command += portname + " | grep ^Path: > " + tempDir.getTempFile();
	int i = system(command.c_str());
	if (logStream.is_open())
	{
		logStream << "DEBUG: command: " << command << endl;
		logStream << "DEBUG: Result: " << i << endl;
	}
	if (!WIFEXITED(i) || WEXITSTATUS(i) < 0 || WEXITSTATUS(i) > 1)
	{
		fprintf(stdout, "\n");
		fprintf(stderr, ("portdowngrade error: can not read from " + portsDir + "/INDEX\n").c_str());
		return false;
	}
	ifstream result(tempDir.getTempFile().c_str());
	string line;
	int portsDirLen = portsDir.length();
	while (!result.fail() && !result.eof())
	{
		line << result;
		if (logStream.is_open())
			logStream << "DEBUG: got line: " << line << endl;
		if (!line.substr(0, 5).compare("Path:"))
		{
			int i;
			for (i = 5; (line[i] == '\t' || line[i] == ' ') && i < line.length(); ++i);
			line.erase(0, i);
			bool valid = false;
			if (line.find(portsDir) == 0 && line.length() > portsDirLen + 1 && line[portsDirLen] == '/' )
			{
				line.erase(0, portsDirLen + 1);
				valid = true;
			}
			if (valid)
				portNames.push_back(line);
			else
				fprintf(stderr, "\nportdowngrade error: can not parse %d\n", line.c_str());
		}
	}
	if (portNames.empty())
	{
		fprintf(stdout, "not found\n");
		return false;
	}
	if (portNames.size() > 1)
	{
		fprintf(stdout, "\n\nFound several matches:\n");
		int j = 1;
		for (vector<string>::iterator i = portNames.begin(); i != portNames.end(); ++i, ++j)
		{
			fprintf(stdout, "%d: %s\n", j, (*i).c_str());
		}
		fprintf(stdout, "\nPlease choose one: ");
		for (;;)
		{
			string s;
			s << cin;
			char * end = 0;
			long l = strtol(s.c_str(), &end, 10);
			if (l != LONG_MIN && l != LONG_MAX && l > 0 && l < j && !*end)
			{
				portname = portNames[l-1];
				fprintf(stdout, "\nDowngrading port: %s\n", portname.c_str());
				break;
			}
			fprintf(stdout, "Type a number between 1 and %d please: ", j-1);
		}
	}
	else
	{
		portname = portNames[0];
		fprintf(stdout, "found: %s\n", portname.c_str());
	}
	return true;
}

bool checkout()
{
	fprintf(stdout, "\nStep 1: Checking out port from CVS repository\n");
	if (serverRootDir.length())
		fprintf(stdout, "CVS root directory: %s\n", serverRootDir.c_str()+1);
	else if (defaultServerRootDir.length())
	{
		serverRootDir = "d" + defaultServerRootDir;
		fprintf(stdout, "CVS root directory: %s\n", defaultServerRootDir.c_str());
	}
	else
	{
		char * c = getenv("CVSROOT");
		if (!c || !*c)
			fprintf(stdout, "No CVS root directory set! Please use the -s option.\n\n");
		else
			fprintf(stdout, "CVS root directory (from CVSROOT environment variable): %s\n", c);
	}
	string command = "cd " + tempDir.getTempPath() + " ; mkdir port ; cvs -Qf" + serverRootDir + " co -d port ports/" + portname;
	int i = system(command.c_str());
	if (logStream.is_open())
	{
		logStream << "DEBUG: command: " << command << endl;
		logStream << "DEBUG: Result: " << i << endl;
	}
	if (!WIFEXITED(i) || WEXITSTATUS(i))
	{
		fprintf(stderr, "portdowngrade error: Can not check out port %s\n", portname.c_str());
		return false;
	}
	return true;
}

bool readPortHistory()
{
	fprintf(stdout, "\nStep 2: Reading the port history from the CVS repository\n");
	string command = "cd " + tempDir.getTempPath() + "/port ; cvs -Qf" + serverRootDir + " log > " + tempDir.getTempFile();
	int i = system(command.c_str());
	if (logStream.is_open())
	{
		logStream << "DEBUG: command: " << command << endl;
		logStream << "DEBUG: Result: " << i << endl;
	}
	if (!WIFEXITED(i) || WEXITSTATUS(i))
	{
		fprintf(stderr, "portdowngrade error: Can not read log of port %s\n", portname.c_str());
		return false;
	}
	return true;
}

bool prefix(const string & s, const char * prefix, string * postfix = 0)
{
	int len = strlen(prefix);
	if (s.length() < len)
		return false;
	for (int i = 0; i < len; i++)
		if (s[i] != prefix[i])
			return false;
	if (postfix)
		*postfix = s.substr(len);
	return true;
}

bool analyzePortHistory()
{
	fprintf(stdout, "\nStep 3: Analyzing the port history from the CVS repository\n");
	ifstream history(tempDir.getTempFile().c_str());
	enum stateEnum {initial, header, logEntry, logComment} state = initial;
	string file, date, comment, line;
	while (!history.fail() && !history.eof())
	{
		line << history;
		switch (state)
		{
			case initial:
				{
					string fileName;
					if (prefix(line, "Working file: ", & fileName))
					{
						state = header;
						file = fileName;
					}
					break;
				}
			case header:
				{
					if (line == "----------------------------")
					{
						state = logEntry;
					}
					break;
				}
			case logEntry:
				{
					string dateLine;
					if (prefix(line, "date: ", & dateLine))
					{
						int pos = dateLine.find(';');
						if (pos > 0)
						{
							date = dateLine.substr(0, pos);
							state = logComment;
						}
						else
							state = header; // this should never happen
					}
					else if (line == "=============================================================================")
					{
						state = initial; // this should never happen
					}
					break;
				}
			case logComment:
				{
					int logEntryFinished = 0;
					if (line == "----------------------------")
					{
						state = logEntry;
						logEntryFinished = 1;
					}
					else if (line == "=============================================================================")
					{
						state = initial;
						logEntryFinished = 2;
					}
					else if (comment.length() > 0 || line.length() > 0)
						 comment += line + "\n";
					if (logEntryFinished)
					{
						CVSCommitEvent event(date, comment, file);
						event.addToVector(CommitEvents);
						comment = date = "";
						if (logEntryFinished == 2)
							file = "";
					}
					break;
				}
		}
	}
	CommitEventsLargerThan sorter;
	stable_sort(CommitEvents.begin(), CommitEvents.end(), sorter);
	CommitEventVectorType result;
	CommitEventVectorType::iterator i = CommitEvents.begin();
	int debugcount = 0, debugcount2 = 0;
	CVSCommitEvent *eventToCopy = 0;
	if (logStream.is_open())
		logStream << "DEBUG: Analyzing commit event " << debugcount << ". Filename: " << (*i).getFileName() << ", date: " << (*i).getDate() << ", unix date: " << (*i).getUnixDate() << ", comment: " << (*i).getComment() << endl;
	while (i != CommitEvents.end())
	{
		string date = (*i).getDate();
		eventToCopy = &(*i);
		CommitEventVectorType::iterator j = i;
		debugcount2 = debugcount;
		while (true)
		{
			i++;
			if (i == CommitEvents.end())
				break;
			else if (logStream.is_open())
				logStream << "DEBUG: Analyzing commit event " << ++debugcount << ". Filename: " << (*i).getFileName() << ", date: " << (*i).getDate() << ", unix date: " << (*i).getUnixDate() << ", comment: " << (*i).getComment() << endl;
			CVSCommitEvent *eventToCopyBackup = eventToCopy;
			int debugcount2Backup = debugcount2;

			if ((*i).isMakefile())
			{
				if (logStream.is_open())
					logStream << "DEBUG: File is Makefile." << endl;
				eventToCopy = &(*i);
				debugcount2 = debugcount;
			}
			if ((*j).getDate() == (*i).getDate())
			{
				if (logStream.is_open())
					logStream << "DEBUG: Times are equal: skip." << endl;
				continue;
			}
			if ((*j).getComment().length() > MinCommentLength && (*j).getComment() == (*i).getComment())
			{
				if (logStream.is_open())
					logStream << "DEBUG: Comments are simmilar: skip." << endl;
				continue;
			}
			if ((*j).diff(*i) < MaxTimeDiff && (*j).getComment() == (*i).getComment())
			{
				if (logStream.is_open())
					logStream << "DEBUG: Times are simmilar: skip." << endl;
				continue;
			}
			if (logStream.is_open())
				logStream << "DEBUG: Add " << debugcount2Backup << " to list." << endl;
			result.push_back(*eventToCopyBackup);
			break;
		}
	}
	if (logStream.is_open())
		logStream << "DEBUG: No more events: Add " << debugcount2 << " to list." << endl;
	result.push_back(*eventToCopy);
	CommitEvents = result;
	return true;
}

string cleanValue(string s)
{
	while (s.length() > 0 && isspace(s[0]))
		s.erase(0, 1);
	int i = 0;
	while (i < s.length() && s[i] != ' ' && s[i] != '#')
		i++;
	return s.substr(0, i);
}

string getPortVersion(const string & date)
{
	string command = "cd " + tempDir.getTempPath() + " ; cvs -Qf" + serverRootDir + " co -d port -D \"" + date + " GMT\" ports/" + portname + "/Makefile";
	int i = system(command.c_str());
	if (logStream.is_open())
	{
		logStream << "DEBUG: command: " << command << endl;
		logStream << "DEBUG: Result: " << i << endl;
	}
	if (!WIFEXITED(i) || WEXITSTATUS(i))
	{
		fprintf(stderr, ("portdowngrade error: can not check out " + portname + "/Makefile\n").c_str());
		return "";
	}
	ifstream makefile((tempDir.getTempPath() + "port/Makefile").c_str());
	string s, version, revision, distname;
	while (!makefile.fail() && !makefile.eof())
	{
		s << makefile;
		if (logStream.is_open())
			logStream << "DEBUG: Makefile line: " << s << endl;
		if (prefix(s, "PORTVERSION=", &s))
			version = s;
		if (prefix(s, "PORTREVISION=", &s))
			revision = s;
		if (prefix(s, "DISTNAME=", &s))
			distname = s;
		if ((version.length() > 0 || distname.length() > 0) && revision.length() > 0)
		{
			break;
		}
		s = "";
	}
	if (version.length() > 0)
		s = cleanValue(version);
	else if (distname.length() > 0)
		s = cleanValue(distname);
	else
	{
		s = "unknown";
		revision = "";
	}
	if (revision.length() > 0)
		s += "_" + cleanValue(revision);
	return s;
}

bool printLog(int startLine, int sizeX, int sizeY, int & pagesize)
{
	set_keypress();
	echo_off();
	pagesize = sizeY-7;
	fprintf(stdout, "\nStep 4: Load port version numbers and present results\n");
	fprintf(stdout, "Keys: <space> : next page                      d : details\n");
	fprintf(stdout, "            p : previous page\n");
	fprintf(stdout, "      <enter> : leave presentation and downdgrade if wanted\n");
	for (int j = 0; j < sizeX-1; j++)
		fprintf(stdout, "=");
	fprintf(stdout, "\nnumber         date         portversion  comment\n");
	int maxCommentLength = sizeX - 42;
	CommitEventVectorType::iterator i = CommitEvents.begin();
	int j = 0;
	for (j = 0; j < startLine && i != CommitEvents.end(); j++)
		i++;
	for (j = 0; j < pagesize && i != CommitEvents.end(); i++, j++)
	{
		if (!(*i).isPortVersionSet())
		{
			CommitEventVectorType::iterator noVersionStart = i;
			CommitEventVectorType::iterator noVersionEnd = i;
			while (noVersionEnd != CommitEvents.end() && !(*noVersionEnd).isMakefile())
				noVersionEnd++;
			string portVersion = getPortVersion((*noVersionEnd).getDate());
			if (portVersion == "")
			{
				restoreTermiosSettings();
				return false;
			}
			for (;;)
			{
				(*noVersionStart).setPortVersion(portVersion);
				if (noVersionStart++ == noVersionEnd)
					break;
				if (noVersionStart == CommitEvents.end())
					break;
			}
		}
		string comment = (*i).getComment().c_str();
		int pos = comment.find('\n');
		if (pos >= 0)
			comment = comment.substr(0, pos);
		int maxLen = maxCommentLength, versionLen = (*i).getPortVersion().length();
		if (versionLen > 11)
			maxLen -= versionLen - 11;
		if (comment.length() > maxLen)
			comment = comment.substr(0, maxLen);
		fprintf(stdout, "%5d  %s  %-11s  %s\n", j + startLine + 1, (*i).getDate().c_str(), (*i).getPortVersion().c_str(), comment.c_str());
		if (kbhit())
		{
			char c = getchar();
			if (c == '\n' || c == 'p' || c == ' ' || c == 'd')
			{
				ungetc(c, stdin);
				break;
			}
		}
	}
	for (; j < pagesize; j++)
		fprintf(stdout, "\n");
	fprintf(stdout, "Total lines: %d. Command:", CommitEvents.size());
	restoreTermiosSettings();
	return true;
}

int getVersionNumber(const char * message, int maxLine, bool zeroAllowed)
{
	for (;;)
	{
		string s;
		fprintf(stdout, "%s", message);
		s << cin;
		fprintf(stdout, "\n");
		long version = -1;
		if (!s.length())
			return -1;
		char * p = 0;
		version = strtol(s.c_str(), & p, 10);
		if (!p || *p || version < (zeroAllowed?0:1) || version > maxLine)
			fprintf(stdout, "Enter nothing or a valid number between %d and %d!\n", (zeroAllowed?0:1), maxLine);
		else
			return (int)version;
	}
}

string showResult()
{
	struct winsize size;
	int size_x = 80, size_y = 24;
	int firstLine = 0, maxLine = CommitEvents.size();
	if (ioctl(STDOUT_FILENO, TIOCGWINSZ, (char *) & size) >= 0)
	{
		size_y = size.ws_row;
		size_x = size.ws_col;
	}
	int pageSize;
	for(; ; )
	{
		if (!printLog(firstLine, size_x, size_y, pageSize))
			return "error";
		for (bool redraw = false; !redraw; )
		{
			set_keypress();
			echo_off();
			char c = getchar();
			restoreTermiosSettings();
			switch (c)
			{
				case '\n':
					{
						int version = getVersionNumber("\nEnter version number to change port to (0: exit): ", maxLine, true);
						redraw = true;
						if (version < 0)
							break;
						if (version > 0)
						{
							CommitEventVectorType::iterator i = CommitEvents.begin();
							for (int j = 1; j < (int)version && i != CommitEvents.end(); j++)
								i++;
							if (i == CommitEvents.end())
							{
								fprintf(stderr, "portdowngrade error: can not find selected version!\n");
								return "";;
							}
							return (*i).getDate();
						}
						return "";
					}
				case 'd':
					{
						int version = getVersionNumber("Enter version number to show details from: ", maxLine, false);
						redraw = true;
						if (version > 0)
						{
							CommitEventVectorType::iterator i = CommitEvents.begin();
							for (int j = 1; j < (int)version && i != CommitEvents.end(); j++)
								i++;
							if (i == CommitEvents.end())
							{
								fprintf(stderr, "portdowngrade error: can not find selected version!\n");
								break;
							}
							fprintf(stdout, "Full comment:\n");
							fprintf(stdout, "%s\n", (*i).getComment().c_str());
							fprintf(stdout, "Press a key");
							set_keypress();
							echo_off();
							char c = getchar();
							restoreTermiosSettings();
							fprintf(stdout, "\n");
						}
						break;
					}
				case ' ':
					{
						if (firstLine + pageSize < maxLine)
						{
							redraw = true;
							firstLine += pageSize;
						}
						break;
					}
				case 'p':
					{
						if (firstLine > 0)
							redraw = true;
						firstLine -= pageSize;
						if (firstLine < 0)
							firstLine = 0;
						break;
					}
			}
		}
	}
	return "";
}

bool synchPort(const string & date)
{
	fprintf(stdout, "\nStep 5: Checking out choosen date of the port from the CVS repository\n");
	string command = "cd " + tempDir.getTempPath() + " ; mkdir synch ; cvs -Qf" + serverRootDir + " co -d synch -D \"" + date + " GMT\" ports/" + portname;
	int i = system(command.c_str());
	if (logStream.is_open())
	{
		logStream << "DEBUG: command: " << command << endl;
		logStream << "DEBUG: Result: " << i << endl;
	}
	if (!WIFEXITED(i) || WEXITSTATUS(i))
	{
		fprintf(stderr, "portdowngrade error: Can not check out port %s\n", portname.c_str());
		return false;
	}
	return true;
}

string addPathes(const string & path1, const string & path2)
{
	if (path1.length() == 0)
		return path2;
	if (path2.length() == 0)
		return path1;
	bool path1HasSlash = path1[path1.length()-1] == '/';
	bool path2HasSlash = path2[0] == '/';
	if ((path1HasSlash && ! path2HasSlash) || (!path1HasSlash && path2HasSlash))
		return path1 + path2;
	if (path1HasSlash && path2HasSlash)
		return path1 + path2.substr(1);
	return path1 + "/" + path2;
}

bool copyFiles(const string & sourcePath, const string & targetPath, const struct stat & fileStat)
{
	mkdir(targetPath.c_str(), fileStat.st_mode);
	chown(targetPath.c_str(), fileStat.st_uid, fileStat.st_gid);
	DIR * dir = opendir(sourcePath.c_str());
	if (!dir)
	{
		fprintf(stderr, "portdowngrade error: Can not temporary directory %s\n", sourcePath.c_str());
		return false;
	}
	struct dirent * entry;
	while (entry = readdir(dir))
	{
		if (strcmp(entry->d_name, ".") && strcmp(entry->d_name, "..") && strcmp(entry->d_name, "CVS"))
		{
			if (entry->d_type == DT_DIR)
			{
				if (!copyFiles(addPathes(sourcePath, entry->d_name), addPathes(targetPath, entry->d_name), fileStat))
					return false;
			}
			else if (entry->d_type == DT_REG)
			{
				string command = "cp -p " + addPathes(sourcePath, entry->d_name) + " " + addPathes(targetPath, entry->d_name);
				int i = system(command.c_str());
				if (logStream.is_open())
				{
					logStream << "DEBUG: command: " << command << endl;
					logStream << "DEBUG: Result: " << i << endl;
				}
				if (!WIFEXITED(i) || WEXITSTATUS(i))
				{
					fprintf(stderr, "portdowngrade error: Can not copy file %s\n", addPathes(targetPath, entry->d_name).c_str());
					return false;
				}
			}
		}
	}
	closedir(dir);
	return true;
}

int copyPortFiles(const string & date)
{
	fprintf(stdout, "\nStep 6: Modifying the port\n");
	fprintf(stdout, "Port: %s\n", portname.c_str());
	fprintf(stdout, "at : %s\n", date.c_str());
	fprintf(stdout, "Type 'yes' to bring the port to the state of the date above\n");
	fprintf(stdout, "or 'no' to exit without changing anything. Note, that this only changes\n");
	fprintf(stdout, "the port, not the installed software! yes or no:");
	for (;;)
	{
		string s;
		s << cin;
		if (s == "yes")
			break;
		if (s == "no")
			return 1;
		fprintf(stdout, "Type 'yes' or 'no' please:");
	}
	struct stat fileStat;
	if (stat((portsDir + "/" + portname).c_str(), &fileStat) < 0)
	{
		fprintf(stderr, "portdowngrade error: Can not access port %s\n", (portsDir + "/" + portname).c_str());
		return -8;
	}
	string command = "rm -rfd " + portsDir + "/" + portname;
	int i = system(command.c_str());
	if (!WIFEXITED(i) || WEXITSTATUS(i))
	{
		fprintf(stderr, "portdowngrade error: Can not delete port %s\n", (portsDir + "/" + portname).c_str());
		return -9;
	}
	return copyFiles(tempDir.getTempPath() + "synch/", portsDir + "/" + portname + "/", fileStat)?0:-9;
}

int main(int argc, char *argv[])
{
	storeTermiosSettings();
	if (!tempDir.tempDirExists())
	{
		fprintf(stderr, "portdowngrade error: can not create temporary directory\n");
		return -3;
	}

	string serverRootDirHelp = "CVS server root directory. If not set, the CVSROOT environment variable is used. Example: \"-s:pserver:anoncvs@anoncvs.FreeBSD.org:/home/ncvs\"";
	string logFileHelp = "log file for debug purposes";
	string maxTimeDiffHelp = "maximum time difference (in seconds) to treat cvs commits with the same comment as belonging together. Default: ";
	string loginHelp = "execute a \"cvs login\" before all other cvs commands. Afterwards, the password for the choosen cvs server should have been stored in the ~/.cvspass file, if it exists, so that this option is no longer required.";
	char tmp[10]; sprintf(tmp, "%d", MaxTimeDiff);
	maxTimeDiffHelp  += tmp;
	
	char    c; // used for argument parsing

	char * arg_serverRootDir = (char *)serverRootDir.c_str();
	char * arg_logFile = 0;
	int arg_maxTime = 0, arg_maxTimeFound = 0, arg_login = 0;
	struct poptOption optionsTable[] =
	{
		{ "server", 's', POPT_ARG_STRING, &arg_serverRootDir, 0, serverRootDirHelp.c_str(), "cvs-server" },
		{ "logfile", 'l', POPT_ARG_STRING, &arg_logFile, 0, logFileHelp.c_str(), "filename" },
		{ "maxTime", 't', POPT_ARG_INT, &arg_maxTime, 1, maxTimeDiffHelp.c_str(), "time" },
		{ "login", 'o', POPT_ARG_NONE, &arg_login, 0, loginHelp.c_str(), NULL },
		POPT_AUTOHELP
		{ NULL, 0, 0, NULL, 0 }
	};

	poptContext optCon = poptGetContext("portdowngrade", argc, (const char **)argv, optionsTable, 0);
	poptSetOtherOptionHelp(optCon, "portname (portname my be prefixed by category/ and postfixed by $)");

	for(;;)
	{
		int (c = poptGetNextOpt(optCon));
		if (c < -1)
		{
			fprintf(stderr, "portdowngrade: invalid argument: %s: %s\n",
				poptBadOption(optCon, POPT_BADOPTION_NOALIAS),
				poptStrerror(c));
			return -1;
		}
		if (c == -1)
			break;
		if (c == 1)
			arg_maxTimeFound = 1;
	}

	const char ** unboundArgs = poptGetArgs(optCon);
	int unboundArgsCount = 0;

	while (unboundArgs && *unboundArgs)
	{
		if (!portname.length())
			portname = *unboundArgs;
		++unboundArgs;
		++unboundArgsCount;
	}
	if (unboundArgsCount != 1)
	{
		poptPrintUsage(optCon, stderr, 0);
		return -1;
	}
	
	if (arg_serverRootDir && *arg_serverRootDir)
		serverRootDir = string("d") + arg_serverRootDir;
	
	if (arg_logFile)
		logStream.open(arg_logFile);
		
	if (arg_maxTimeFound)
	{
		if (arg_maxTime < 0)
		{
			fprintf(stderr, "portdowngrade: maxTime must be non-negative\n");
			return -1;
		}
		MaxTimeDiff = arg_maxTime;
	}

	fprintf(stdout, "\nportdowngrade %d.%d by Heiner Eichmann\n", VERSION_HI, VERSION_LO);
	fprintf(stdout, "Please note, that nothing is changed in the ports tree\n");
	fprintf(stdout, "unless it is explicitly permitted in step 6!\n\n");
	
	if (!completePortPath(portname))
		return -2;

	if (arg_login)
	{
		string rootDir = serverRootDir;
		if (!rootDir.length() && defaultServerRootDir.length())
			rootDir = "d" + defaultServerRootDir;
		if (!rootDir.length())
		{
			char * c = getenv("CVSROOT");
			if (!c || !*c)
			{
				fprintf(stderr, "No CVS root directory set!\n");
				return -1;
			}
			fprintf(stdout, "login to CVS root directory from CVSROOT environment variable: %s\n", c);
			rootDir = "d";
			rootDir += c;
		}
		string command = "cvs -" + rootDir + " login";
		int i = system(command.c_str());
		if (logStream.is_open())
		{
			logStream << "DEBUG: command: " << command << endl;
			logStream << "DEBUG: Result: " << i << endl;
		}
		if (!WIFEXITED(i) || WEXITSTATUS(i))
		{
			fprintf(stderr, "portdowngrade error: Login failed\n");
			return false;
		}
	}
	int result = 0;
	for(;;)
	{
		if (!checkout())
		{
			result = -4;
			break;
		}
		if (!readPortHistory())
		{
			result = -5;
			break;
		}
		if (!analyzePortHistory())
		{
			result = -5;
			break;
		}
		string selectedVersion = showResult();
		if (selectedVersion == "error")
		{
			result = -6;
			break;
		}
		if (!selectedVersion.length())
			break;
		if (!synchPort(selectedVersion))
		{
			result = -7;
			break;
		}
		result = copyPortFiles(selectedVersion);
		break;
	}
	if (result == 0)
	{
		int pos = (int)portname.find('/');
		if (pos > 0)
			portname = portname.substr(pos+1);
		fprintf(stdout, "\nThe port has been set to the selected version. Install it if you wish.\n");
		fprintf(stdout, "If you have portupgrade installed, you should run\n\n");
		fprintf(stdout, "portsdb -Uu\n\n");
		fprintf(stdout, "now, to see the changes in the ports database. In any case\n\n");
		fprintf(stdout, "portupgrade -f %s\n\n", portname.c_str());
		fprintf(stdout, "will install the changed port. Note: if you run cvsup, the port\n");
		fprintf(stdout, "is changed back to the choosen label!\n\n");
	}
	else if (result < 0)
	{
		fprintf(stdout, "\nAn error had occured. If this was an cvs error, try another cvs server.\n");
		fprintf(stdout, "If this was an cvs login error, try the -o option.\n");
		fprintf(stdout, "\nSee http://www.freebsd.org/doc/en_US.ISO8859-1/books/handbook/anoncvs.html for details.\n");
	}
	return result;
}

